#include <gst/editor/editor.h>

/* class functions */
static void 	gst_editor_connection_class_init	(GstEditorConnectionClass *klass);
static void 	gst_editor_connection_init		(GstEditorConnection *connection);

static void	gst_editor_connection_set_property	(GObject *object, guint prop_id, 
                                                         const GValue *value, GParamSpec *pspec);
static void	gst_editor_connection_get_property	(GObject *object, guint prop_id, 
                                                         GValue *value, GParamSpec *pspec);
static void 	gst_editor_connection_realize		(GnomeCanvasItem *citem);
static void 	gst_editor_connection_resize		(GstEditorConnection *connection);

/* callbacks from gstreamer */
static void	on_pad_disconnect			(GstPad *peer,
                                                         GstPad *pad,
                                                         GstEditorConnection *connection);

/* callbacks from editor pads */
static void	on_editor_pad_position_changed		(GstEditorPad *pad,
                                                         GstEditorConnection *connection);


enum {
  PROP_0,
  PROP_X,
  PROP_Y,
  PROP_X1,
  PROP_Y1,
  PROP_X2,
  PROP_Y2,
  PROP_SRCPAD,
  PROP_SINKPAD,
  PROP_GHOST,
};

enum {
  LAST_SIGNAL
};

static GObjectClass *parent_class;
/* static guint gst_editor_connection_signals[LAST_SIGNAL] = { 0 }; */

GType 
gst_editor_connection_get_type (void) 
{
  static GType connection_type = 0;

  if (!connection_type) {
    static const GTypeInfo connection_info = {
      sizeof (GstEditorConnectionClass),
      (GBaseInitFunc) NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc) gst_editor_connection_class_init,
      NULL,
      NULL,
      sizeof (GstEditorConnection),
      0,
      (GInstanceInitFunc) gst_editor_connection_init,
    };
      
    connection_type = g_type_register_static (gnome_canvas_line_get_type (), "GstEditorConnection", &connection_info, 0);
  }
  return connection_type;
}

static void 
gst_editor_connection_class_init (GstEditorConnectionClass *klass) 
{
  GObjectClass *object_class;

  object_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_ref (gnome_canvas_line_get_type ());

  object_class->set_property = gst_editor_connection_set_property;
  object_class->get_property = gst_editor_connection_get_property;

  g_object_class_install_property (object_class, PROP_X,
       g_param_spec_double ("x", "x", "x",
			    -G_MAXDOUBLE, G_MAXDOUBLE, 0, G_PARAM_READWRITE));
  g_object_class_install_property (object_class, PROP_Y,
       g_param_spec_double ("y", "y", "y",
			    -G_MAXDOUBLE, G_MAXDOUBLE, 0, G_PARAM_READWRITE));
  g_object_class_install_property (object_class, PROP_X1,
       g_param_spec_double ("x1", "x1", "x1",
			    -G_MAXDOUBLE, G_MAXDOUBLE, 0, G_PARAM_READABLE));
  g_object_class_install_property (object_class, PROP_Y1,
       g_param_spec_double ("y1", "y1", "y1",
			    -G_MAXDOUBLE, G_MAXDOUBLE, 0, G_PARAM_READABLE));
  g_object_class_install_property (object_class, PROP_X2,
       g_param_spec_double ("x2", "x2", "x2",
			    -G_MAXDOUBLE, G_MAXDOUBLE, 0, G_PARAM_READABLE));
  g_object_class_install_property (object_class, PROP_Y2,
       g_param_spec_double ("y2", "y2", "y2",
			    -G_MAXDOUBLE, G_MAXDOUBLE, 0, G_PARAM_READABLE));
  g_object_class_install_property (object_class, PROP_SRCPAD,
       g_param_spec_object ("src-pad", "src-pad", "src-pad",
                             gst_editor_pad_get_type (), G_PARAM_READWRITE));
  g_object_class_install_property (object_class, PROP_SINKPAD,
       g_param_spec_object ("sink-pad", "sink-pad", "sink-pad",
                             gst_editor_pad_get_type (), G_PARAM_READWRITE));
  g_object_class_install_property (object_class, PROP_GHOST,
       g_param_spec_boolean ("ghost", "ghost", "ghost",
                             FALSE, G_PARAM_READWRITE));

  GNOME_CANVAS_ITEM_CLASS (klass)->realize = gst_editor_connection_realize;
}

static void 
gst_editor_connection_init (GstEditorConnection *connection) 
{
  connection->points = gnome_canvas_points_new (2);

  gnome_canvas_item_raise ((GnomeCanvasItem*)connection, 10);
}

static void 
gst_editor_connection_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec) 
{
  GstEditorPad *srcpad, *sinkpad;
  GstEditorConnection *connection = GST_EDITOR_CONNECTION (object);

  switch (prop_id) {
    case PROP_X:
      if (connection->srcpad && connection->sinkpad)
        g_warning ("Settting connection drag x without having one unset pad");
      
      connection->dragging = TRUE;
      connection->x = g_value_get_double (value);
      break;

    case PROP_Y:
      if (connection->srcpad && connection->sinkpad)
        g_warning ("Settting connection drag y without having one unset pad");
      
      connection->dragging = TRUE;
      connection->y = g_value_get_double (value);
      break;

    case PROP_SRCPAD:
      srcpad  = (GstEditorPad*)connection->srcpad;
      sinkpad = (GstEditorPad*)connection->sinkpad;

      if (srcpad) {
        if (connection->ghost)
          srcpad->ghostconnection = NULL;
        else
          srcpad->connection = NULL;
      }

      srcpad = (GstEditorPad*)g_value_get_object (value);
      
      if (srcpad) {
        if (connection->ghost)
          srcpad->ghostconnection = connection;
        else
          srcpad->connection = connection;

        if (sinkpad)
          connection->dragging = FALSE;
      }

      connection->srcpad = (GstEditorItem*)srcpad;
      break;

    case PROP_SINKPAD:
      srcpad  = (GstEditorPad*)connection->srcpad;
      sinkpad = (GstEditorPad*)connection->sinkpad;

      if (sinkpad) {
        if (connection->ghost)
          sinkpad->ghostconnection = NULL;
        else
          sinkpad->connection = NULL;
      }

      sinkpad = (GstEditorPad*)g_value_get_object (value);
      
      if (sinkpad) {
        if (connection->ghost)
          sinkpad->ghostconnection = connection;
        else
          sinkpad->connection = connection;

        if (srcpad)
          connection->dragging = FALSE;
      }

      connection->sinkpad = (GstEditorItem*)sinkpad;
      break;

    case PROP_GHOST:
      connection->ghost = g_value_get_boolean (value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }

  /* check if we're ready to resize */
  if (((connection->srcpad || connection->sinkpad) && connection->dragging) ||
      (connection->srcpad && connection->sinkpad))
    gst_editor_connection_resize (connection);
}

static void 
gst_editor_connection_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec) 
{
  GstEditorConnection *connection = GST_EDITOR_CONNECTION (object);
  gdouble d = 0.0, blah = 0.0;

  switch (prop_id) {
    case PROP_X:
      g_value_set_double (value, connection->x);
      break;

    case PROP_Y:
      g_value_set_double (value, connection->y);
      break;

    case PROP_X1:
      if (connection->srcpad) {
        g_object_get (connection->srcpad, "x", &d, NULL);
        d += connection->srcpad->width;
        gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (connection->srcpad)->parent, &d, &blah);
      } else if (connection->dragging) {
        d = connection->x;
      } else {
        g_warning ("no src pad");
      }
      g_value_set_double (value, d);
      break;

    case PROP_X2:
      if (connection->sinkpad) {
        g_object_get (connection->sinkpad, "x", &d, NULL);
        gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (connection->sinkpad)->parent, &d, &blah);
      } else if (connection->dragging) {
        d = connection->x;
      } else {
        g_warning ("no sink pad");
      }
      
      g_value_set_double (value, d);
      break;

    case PROP_Y1:
      if (connection->srcpad) {
        g_object_get (connection->srcpad, "y", &d, NULL);
        d += connection->srcpad->height / 2;
        gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (connection->srcpad)->parent, &blah, &d);
      } else if (connection->dragging) {
        d = connection->y;
      } else {
        g_warning ("no src pad");
      }
      g_value_set_double (value, d);
      break;

    case PROP_Y2:
      if (connection->sinkpad) {
        g_object_get (connection->sinkpad, "y", &d, NULL);
        d += connection->sinkpad->height / 2;
        gnome_canvas_item_i2w (GNOME_CANVAS_ITEM (connection->sinkpad)->parent, &blah, &d);
      } else if (connection->dragging) {
        d = connection->y;
      } else {
        g_warning ("no sink pad");
      }
      g_value_set_double (value, d);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


static void 
gst_editor_connection_realize (GnomeCanvasItem *citem) 
{
  GstEditorConnection *connection = GST_EDITOR_CONNECTION (citem);
         
  connection->points->coords[0] = 0.0;
  connection->points->coords[1] = 0.0;
  connection->points->coords[2] = 0.0;
  connection->points->coords[3] = 0.0;

  /* we need to be realized before setting properties */
  if (GNOME_CANVAS_ITEM_CLASS (parent_class)->realize)
    GNOME_CANVAS_ITEM_CLASS (parent_class)->realize (citem);

  /* see gnome-canvas-line.h for the docs */
  gnome_canvas_item_set (citem,
                         "points", connection->points,
                         "width-units", 2.0,
                         "line-style", GDK_LINE_ON_OFF_DASH,
                         "first-arrowhead", TRUE,
                         "arrow-shape-a", 5.0,
                         "arrow-shape-b", 5.0,
                         "arrow-shape-c", 5.0,
                         NULL);
}

static void 
gst_editor_connection_resize (GstEditorConnection *connection) 
{
  gdouble x1,y1,x2,y2;

  g_object_get (connection, "x1", &x1, "y1", &y1, "x2", &x2, "y2", &y2, NULL);
  gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (connection)->parent, &x1, &y1);
  gnome_canvas_item_w2i (GNOME_CANVAS_ITEM (connection)->parent, &x2, &y2);

  /* we do this in reverse so that with dot-dash lines it gives the illusion of
     pulling out a rope from the element */
  connection->points->coords[2] = x1;
  connection->points->coords[3] = y1;
  connection->points->coords[0] = x2;
  connection->points->coords[1] = y2;
  gnome_canvas_item_set (GNOME_CANVAS_ITEM (connection),
                         "points", connection->points, NULL);
}

static void
on_pad_disconnect (GstPad *peer, GstPad *pad, GstEditorConnection *connection)
{
  GstEditorBin *srcbin, *sinkbin;

  g_message ("disconnect pad signal (%s:%s from %s:%s)",
             GST_DEBUG_PAD_NAME (pad), GST_DEBUG_PAD_NAME (peer));

  g_signal_handlers_disconnect_by_func (connection->srcpad,
                                        on_editor_pad_position_changed, connection);
  g_signal_handlers_disconnect_by_func (connection->sinkpad,
                                        on_editor_pad_position_changed, connection);
  g_signal_handlers_disconnect_by_func (pad, on_pad_disconnect, connection);

  srcbin  = GST_EDITOR_BIN (GNOME_CANVAS_ITEM (connection->srcpad)->parent->parent);
  sinkbin = GST_EDITOR_BIN (GNOME_CANVAS_ITEM (connection->sinkpad)->parent->parent);
        
  sinkbin->connections = g_list_remove (sinkbin->connections, connection);
  if (sinkbin != srcbin)
    srcbin->connections = g_list_remove (srcbin->connections, connection);

  GST_EDITOR_PAD (connection->srcpad)->connection = NULL;
  GST_EDITOR_PAD (connection->sinkpad)->connection = NULL;
  connection->srcpad = NULL;
  connection->sinkpad = NULL;
  /* i have bad luck with actually killing the GCI's */
  gnome_canvas_item_hide (GNOME_CANVAS_ITEM (connection));
}

static void
on_editor_pad_position_changed (GstEditorPad *pad, GstEditorConnection *connection)
{
  g_return_if_fail (GST_IS_EDITOR_CONNECTION (connection));

  gst_editor_connection_resize (connection);
}

gboolean 
gst_editor_connection_connect (GstEditorConnection *connection)
{
  GObject *src, *sink;
  GstPad *srcpad = NULL, *sinkpad = NULL;
  
  g_return_val_if_fail (GST_IS_EDITOR_CONNECTION (connection), FALSE);

  if (!connection->srcpad || !connection->sinkpad)
    goto error;

  src  = (GObject*) GST_EDITOR_ITEM (connection->srcpad)->object;
  sink = (GObject*) GST_EDITOR_ITEM (connection->sinkpad)->object;

  if (!GST_EDITOR_PAD (connection->srcpad)->istemplate) {
    if (!GST_EDITOR_PAD (connection->sinkpad)->istemplate) {
      if (GST_PAD_PEER (src) || GST_PAD_PEER (sink)) {
        if (GST_PAD_PEER (src) != (GstPad*)sink) {
          g_warning ("The src pad is connected, but not to the sink pad");
          goto error;
        }
        if (GST_PAD_PEER (sink) != (GstPad*)src) {
          g_warning ("The sink pad is connected, but not to the src pad");
          goto error;
        }
        srcpad = GST_PAD (src);
        sinkpad = GST_PAD (sink);
        goto connected;
        /* yay goto */
      }
    } else if (GST_PAD_PEER (src)) {
      /* the to pad is a template */
      g_warning ("The src pad is connected, but not to the sink pad");
      goto error;
    }
  } else if (!GST_EDITOR_PAD (connection->sinkpad)->istemplate && GST_PAD_PEER (sink)) {
    /* from pad is a template */
    g_warning ("The sink pad is connected, but not to the src pad");
    goto error;
  }
  
  if (connection->ghost) {
      gnome_canvas_item_set (GNOME_CANVAS_ITEM (connection), "fill-color",
                             "grey70", NULL);
      goto connected;
  } else {
    if (!GST_EDITOR_PAD (connection->srcpad)->istemplate) {
      srcpad = GST_PAD (src);
    } else if (GST_IS_EDITOR_PAD_REQUEST (connection->srcpad)) {
      srcpad = gst_element_get_request_pad ((GstElement*)GST_EDITOR_ITEM (GNOME_CANVAS_ITEM (connection->srcpad)->parent)->object,
                                            GST_PAD_TEMPLATE (src)->name_template);
      /* the new_pad signal will cause a new pad to made automagically in the
         element */
      gnome_canvas_item_set (GNOME_CANVAS_ITEM (connection), "src-pad",
                             gst_editor_item_get ((GstObject*)srcpad), NULL);
    } else {
      goto error;
    }
    
    if (!srcpad)
      goto error;
    
    if (!GST_EDITOR_PAD (connection->sinkpad)->istemplate) {
      sinkpad = GST_PAD (sink);
    } else if (GST_IS_EDITOR_PAD_REQUEST (connection->sinkpad)) {
      sinkpad = gst_element_get_request_pad ((GstElement*)GST_EDITOR_ITEM (GNOME_CANVAS_ITEM (connection->sinkpad)->parent)->object,
                                             GST_PAD_TEMPLATE (sink)->name_template);
      gnome_canvas_item_set (GNOME_CANVAS_ITEM (connection), "sink-pad",
                             gst_editor_item_get ((GstObject*)sinkpad), NULL);
    } else {
      goto error;
    }
    
    if (!sinkpad)
      goto error;


    if (gst_pad_connect (srcpad, sinkpad)) {
      GstEditorBin *srcbin, *sinkbin;

    connected:

      srcbin = sinkbin = NULL;

      gnome_canvas_item_set (GNOME_CANVAS_ITEM (connection), "line-style",
                             GDK_LINE_SOLID, NULL);

      g_signal_connect (connection->srcpad, "position-changed",
                        G_CALLBACK (on_editor_pad_position_changed), G_OBJECT (connection));
      g_signal_connect (connection->sinkpad, "position-changed",
                        G_CALLBACK (on_editor_pad_position_changed), G_OBJECT (connection));

      if (!connection->ghost)
        g_signal_connect (srcpad, "disconnected",
                          G_CALLBACK (on_pad_disconnect), connection);

      if (GST_IS_EDITOR_BIN (GNOME_CANVAS_ITEM (connection->srcpad)->parent->parent))
        srcbin  = GST_EDITOR_BIN (GNOME_CANVAS_ITEM (connection->srcpad)->parent->parent);
      if (GST_IS_EDITOR_BIN (GNOME_CANVAS_ITEM (connection->sinkpad)->parent->parent))
        sinkbin = GST_EDITOR_BIN (GNOME_CANVAS_ITEM (connection->sinkpad)->parent->parent);
        
      if (sinkbin)
        sinkbin->connections = g_list_prepend (sinkbin->connections, connection);
      if (srcbin && sinkbin != srcbin)
        srcbin->connections = g_list_prepend (srcbin->connections, connection);

      return TRUE;
    }
  }

error:
  g_message ("could not connect");

  if (connection->srcpad)
    GST_EDITOR_PAD (connection->srcpad)->connection = NULL;
  if (connection->sinkpad)
    GST_EDITOR_PAD (connection->sinkpad)->connection = NULL;
  return FALSE;
}
